import json
import logging
from datetime import datetime
from datetime import timedelta
from pathlib import Path

import pytz
from twisted.internet.task import LoopingCall
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.DeferUtil import vortexLogFailure
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.UserLoggedIn import UserLoggedIn
from peek_plugin_base.LoopingCallUtil import peekCatchErrbackWithLogger

logger = logging.getLogger(__name__)


@addTupleType
class UserMetricsTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "UserMetricsTuple"

    lastUpdateDate: datetime = TupleField()

    # User metrics
    totalUsers: int = TupleField()
    loginsLast365Days: int = TupleField()
    loginsLast30Days: int = TupleField()
    loginsLast7Days: int = TupleField()
    loginsLast24Hours: int = TupleField()
    totalLoggedInDevices: int = TupleField()
    totalFieldLogins: int = TupleField()


class MetricsController:
    LOOPING_CALL_PERIOD = 30.0 * 60  # 30 minutes

    def __init__(self, writeDir: Path, ormSessionCreator):
        self._writeDir = writeDir
        self._ormSessionCreator = ormSessionCreator
        self._loopingCall = LoopingCall(
            peekCatchErrbackWithLogger(logger)(self.writeMetrics)
        )

    def start(self):
        d = self._loopingCall.start(self.LOOPING_CALL_PERIOD, now=True)
        d.addErrback(vortexLogFailure, logger, consumeError=True)

    def shutdown(self):
        if self._loopingCall is not None:
            self._loopingCall.stop()
            self._loopingCall = None

    def _maybeNaive(self, dt: datetime | None) -> datetime | None:
        """Convert datetime to UTC if it has a timezone, or assume UTC if naive.

        Args:
            dt: The datetime to convert

        Returns:
            UTC datetime or None if input is None
        """
        if dt is None:
            return None

        if dt.tzinfo is None:
            return dt.replace(tzinfo=pytz.UTC)

        return dt.astimezone(pytz.UTC)

    @deferToThreadWrapWithLogger(logger)
    def writeMetrics(self) -> None:
        """Collect and write system metrics to a JSON file.

        Raises:
            OSError: If there are file permission issues
            psutil.Error: If there are errors collecting system metrics
        """
        try:
            metrics = self._collectMetrics()
            jsonFilePath = self._writeDir / "metrics.json"

            # Convert datetime to string for JSON serialization
            metricsDict = metrics.tupleToRestfulJsonDict()

            with jsonFilePath.open("w") as f:
                json.dump(metricsDict, f, indent=4)

        except Exception as e:
            logger.exception(f"Error writing metrics: {e}")
            raise

    def _collectUserMetrics(self, metrics: UserMetricsTuple) -> None:
        """Collect user-related metrics."""
        ormSession = self._ormSessionCreator()
        try:
            now = datetime.now(pytz.utc)

            # Time thresholds
            oneYearAgo = now - timedelta(days=365)
            oneMonthAgo = now - timedelta(days=30)
            oneWeekAgo = now - timedelta(days=7)
            todayStart = now - timedelta(days=1)

            # Get all users and their last login dates
            users = list(ormSession.query(InternalUserTuple))
            metrics.totalUsers = len(users)

            # Calculate login metrics based on lastLoginDate
            metrics.loginsLast365Days = sum(
                1
                for user in users
                if self._maybeNaive(user.lastLoginDate)
                and self._maybeNaive(user.lastLoginDate) >= oneYearAgo
            )
            metrics.loginsLast30Days = sum(
                1
                for user in users
                if self._maybeNaive(user.lastLoginDate)
                and self._maybeNaive(user.lastLoginDate) >= oneMonthAgo
            )
            metrics.loginsLast7Days = sum(
                1
                for user in users
                if self._maybeNaive(user.lastLoginDate)
                and self._maybeNaive(user.lastLoginDate) >= oneWeekAgo
            )
            metrics.loginsLast24Hours = sum(
                1
                for user in users
                if self._maybeNaive(user.lastLoginDate)
                and self._maybeNaive(user.lastLoginDate) >= todayStart
            )

            # Get current logged in sessions
            loggedInSessions = list(ormSession.query(UserLoggedIn))

            # Count total logged in devices (all active sessions)
            metrics.totalLoggedInDevices = len(loggedInSessions)

            # Count field logins (only one allowed per user)
            metrics.totalFieldLogins = sum(
                1 for session in loggedInSessions if session.isFieldLogin
            )

        finally:
            ormSession.close()

    def _collectMetrics(self) -> UserMetricsTuple:
        """Collect all user metrics and return as a UserMetricsTuple."""
        metrics = UserMetricsTuple(lastUpdateDate=datetime.now(pytz.utc))
        self._collectUserMetrics(metrics)
        return metrics



"""added foreign key to user logged in

Peek Plugin Database Migration Script

Revision ID: 1176bc54abf3
Revises: 7e3f668edf0e
Create Date: 2021-03-22 10:35:44.330884

"""

# revision identifiers, used by Alembic.
revision = "1176bc54abf3"
down_revision = "7e3f668edf0e"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint("uq_username", "UserLoggedIn", ["userName"])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("uq_username", "UserLoggedIn")
    # ### end Alembic commands ###



"""removed unique logged in user

Peek Plugin Database Migration Script

Revision ID: 5e33c6b788fd
Revises: 3fd59734e444
Create Date: 2019-07-19 15:00:09.318828

"""

# revision identifiers, used by Alembic.
revision = "5e33c6b788fd"
down_revision = "3fd59734e444"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "UserLoggedIn_userName_key",
        "UserLoggedIn",
        schema="core_user",
        type_="unique",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint(
        "UserLoggedIn_userName_key",
        "UserLoggedIn",
        ["userName"],
        schema="core_user",
    )
    # ### end Alembic commands ###



from typing import Dict, List

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user.tuples.UserDetailTuple import UserDetailTuple
from vortex.Tuple import addTupleType, Tuple, TupleField


@addTupleType
class UserLoginResponseTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "UserLoginResponseTuple"
    userName: str = TupleField()
    userToken: str = TupleField()
    deviceToken: str = TupleField()
    deviceDescription: str = TupleField()
    vehicleId: str = TupleField()

    userDetail: UserDetailTuple = TupleField()

    succeeded: bool = TupleField(True)

    #: A list of accepted warning keys
    # If any server side warnings occur and they are in this list then the logon
    # continues.
    acceptedWarningKeys: List[str] = TupleField()

    #: A dict of warnings from a failed logon action.
    # key = a unique key for this warning
    # value = the description of the warning for the user
    warnings: Dict[str, str] = TupleField({})

    def setFailed(self) -> None:
        self.succeeded = False

    def addWarning(self, key: str, displayText: str):
        self.warnings[key] = displayText



import logging
from typing import List, Optional

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField

logger = logging.getLogger(__name__)


@addTupleType
class ImportInternalUserTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "ImportInternalUserTuple"

    #: The unique name of the user
    userName: str = TupleField()

    #: The unique uuid of the user
    userUuid: str = TupleField()

    #: The nice name of the user
    userTitle: str = TupleField()

    #: The password for the user, null if it's managed in the peek_user_plugin
    password: Optional[str] = TupleField()

    #: Group Keys
    groupKeys: Optional[List[str]] = TupleField()

    #: The mobile phone number of the user
    mobile: Optional[str] = TupleField()

    #: The email address of the user
    email: Optional[str] = TupleField()

    #: The user creation stream - arbitrary string
    importSource: str = TupleField()

    #: The user authentication target
    #  INTERNAL or LDAP from `UserAuthTargetEnum`
    authenticationTarget: str = TupleField()



""" 
 *  Copyright Synerty Pty Ltd 2016
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

import logging

from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase
from peek_core_user.tuples.UserLoggedInInfoTuple import UserLoggedInInfoTuple

logger = logging.getLogger(__name__)


@addTupleType
class UserLoggedIn(DeclarativeBase, Tuple):
    """UserLoggedIn

    This table stores which users are logged into which devices

    """

    __tupleType__ = userPluginTuplePrefix + "UserLoggedIn"
    __tablename__ = "UserLoggedIn"

    id = Column(Integer, primary_key=True, autoincrement=True)
    loggedInDateTime = Column(DateTime(True), nullable=False)
    userName = Column(String, nullable=False)
    userTitle = Column(String)
    userKey = Column(String, nullable=False)
    userUuid = Column(String, nullable=False)
    deviceToken = Column(String(100), unique=True, nullable=False)
    vehicle = Column(String)
    isFieldLogin = Column(Boolean)
    loggedInWithGroups: str = Column(String)

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)

    def toTuple(self):
        return UserLoggedInInfoTuple(
            loggedInDateTime=self.loggedInDateTime,
            userName=self.userName,
            userTitle=self.userTitle,
            deviceToken=self.deviceToken,
            vehicle=self.vehicle,
            isFieldLogin=self.isFieldLogin,
            loggedInWithGroups=self.loggedInWithGroups,
        )



import logging
from abc import ABCMeta
from abc import abstractmethod
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple

from sqlalchemy import Row
from sqlalchemy import func
from sqlalchemy import select
from sqlalchemy.orm.exc import NoResultFound
from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger

from peek_core_user._private.storage.Setting import (
    INTERNAL_AUTH_ENABLED_FOR_ADMIN,
)
from peek_core_user._private.storage.Setting import (
    INTERNAL_AUTH_ENABLED_FOR_FIELD,
)
from peek_core_user._private.storage.Setting import (
    INTERNAL_AUTH_ENABLED_FOR_OFFICE,
)
from peek_core_user._private.storage.Setting import LDAP_AUTH_ENABLED
from peek_core_user._private.storage.Setting import globalSetting
from peek_core_user._private.storage.InternalUserPassword import (
    InternalUserPassword,
)
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_plugin_base.storage.DbConnection import DbSessionCreator

logger = logging.getLogger(__name__)


class AuthSettings(NamedTuple):
    internalFieldEnabled: bool
    internalOfficeEnabled: bool
    internalAdminEnabled: bool
    ldapAuthEnabled: bool


class AuthABC(metaclass=ABCMeta):
    FOR_SERVICE_ADMIN = 1
    FOR_SERVICE_OFFICE = 2
    FOR_SERVICE_FIELD = 3
    FOR_SERVICE_FIELD_OTP = 4

    def __init__(self, dbSessionCreator: DbSessionCreator):
        self._dbSessionCreator = dbSessionCreator

    @abstractmethod
    def checkPassAsync(
        self, userName: str, password: str, forService: int
    ) -> Tuple[List[str], InternalUserTuple]:
        raise NotImplementedError()

    @deferToThreadWrapWithLogger(logger)
    def getInternalUser(
        self, userName, raiseNotLoggedInException=True
    ) -> Optional[InternalUserTuple]:
        ormSession = self._dbSessionCreator()
        try:
            authenticatedUsers = (
                ormSession.query(InternalUserTuple)
                .filter(InternalUserTuple.userName == userName)
                .all()
            )
            ormSession.expunge_all()

            if len(authenticatedUsers) == 0:
                if raiseNotLoggedInException:
                    raise NoResultFound()
                else:
                    return None

            if len(authenticatedUsers) != 1:
                if raiseNotLoggedInException:
                    raise Exception("Too many users found")
                else:
                    return None

            return authenticatedUsers[0]

            # No commit needed, we only query
        finally:
            ormSession.close()

    @deferToThreadWrapWithLogger(logger)
    def getInternalUserAndPassword(
        self, userName
    ) -> Deferred[Row[InternalUserTuple, InternalUserPassword]]:
        ormSession = self._dbSessionCreator()
        try:
            statement = (
                select(InternalUserTuple, InternalUserPassword)
                .join(
                    InternalUserPassword, isouter=True
                )  # effectively `LEFT JOIN`
                .filter(
                    func.lower(InternalUserTuple.userName) == userName.lower()
                )
            )
            authenticatedUserPasswords = ormSession.execute(
                statement
            ).fetchall()
            ormSession.expunge_all()

            if len(authenticatedUserPasswords) == 0:
                raise NoResultFound()

            if len(authenticatedUserPasswords) != 1:
                raise Exception("Too many users found")

            return authenticatedUserPasswords[0]

            # No commit needed, we only query
        finally:
            ormSession.close()

    @deferToThreadWrapWithLogger(logger)
    def getAuthSettings(self) -> AuthSettings:
        """
        Returns Deferred[UserLogoutResponseTuple]
        """

        ormSession = self._dbSessionCreator()
        try:
            settings = globalSetting(ormSession)
            return AuthSettings(
                internalFieldEnabled=settings[INTERNAL_AUTH_ENABLED_FOR_FIELD],
                internalOfficeEnabled=settings[
                    INTERNAL_AUTH_ENABLED_FOR_OFFICE
                ],
                internalAdminEnabled=settings[INTERNAL_AUTH_ENABLED_FOR_ADMIN],
                ldapAuthEnabled=settings[LDAP_AUTH_ENABLED],
            )

        finally:
            ormSession.close()

    @deferToThreadWrapWithLogger(logger)
    def getInternalUserAndOtp(self, userName) -> Deferred[InternalUserTuple]:
        ormSession = self._dbSessionCreator()
        try:
            authenticatedUserOtps = (
                ormSession.query(InternalUserTuple)
                .filter(
                    func.lower(InternalUserTuple.userName) == userName.lower()
                )
                .all()
            )
            ormSession.expunge_all()

            if len(authenticatedUserOtps) == 0:
                raise NoResultFound()

            if len(authenticatedUserOtps) != 1:
                raise Exception("Too many users found")

            return authenticatedUserOtps[0]

            # No commit needed, we only query
        finally:
            ormSession.close()

    @classmethod
    def getForServiceName(cls, forService: int):
        if forService == cls.FOR_SERVICE_ADMIN:
            return "admin"
        elif forService == cls.FOR_SERVICE_OFFICE:
            return "office"
        elif forService == cls.FOR_SERVICE_FIELD:
            return "field"
        elif forService == cls.FOR_SERVICE_FIELD_OTP:
            return "field otp"
        return "unknown"



import logging
from typing import List
from typing import Optional

from sqlalchemy import or_
from sqlalchemy.orm import subqueryload
from sqlalchemy.orm.exc import NoResultFound
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.DeferUtil import noMainThread

from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_user._private.storage.InternalGroupTuple import (
    InternalGroupTuple,
)
from peek_core_user._private.storage.InternalUserGroupTuple import (
    InternalUserGroupTuple,
)
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user._private.storage.UserLoggedIn import UserLoggedIn
from peek_core_user.server.UserInfoApiABC import UserInfoApiABC
from peek_core_user.tuples.DeviceWithUserDetails import (
    DeviceWithUserDetailsTuple,
)
from peek_core_user.tuples.GroupDetailTuple import GroupDetailTuple
from peek_core_user.tuples.UserDetailTuple import UserDetailTuple

logger = logging.getLogger(__name__)


class UserInfoApi(UserInfoApiABC):
    #: A reference to the core device plugins API
    _deviceApi: DeviceApiABC

    _userCopyFields = set(UserDetailTuple.tupleFieldNames()) & set(
        InternalUserTuple.tupleFieldNames()
    )

    _groupCopyFields = set(GroupDetailTuple.tupleFieldNames()) & set(
        InternalGroupTuple.tupleFieldNames()
    )

    def __init__(self):
        self._deviceApi = None
        self._dbSessionCreator = None

    def setStartValues(self, deviceApi: DeviceApiABC, dbSessionCreator):
        self._deviceApi = deviceApi
        self._dbSessionCreator = dbSessionCreator

    def shutdown(self):
        self._deviceApi = None
        self._dbSessionCreator = None

    @deferToThreadWrapWithLogger(logger)
    def user(self, userName: str):
        """
        :returns a Deferred firing with Optional[UserDetailTuple]]
        """
        return self.userBlocking(userName)

    def userBlocking(
        self, userName, ormSession=None
    ) -> Optional[UserDetailTuple]:
        filter_ = InternalUserTuple.userName == userName
        return self._userInfoByFilter(filter_, ormSession)

    def _userInfoByFilter(
        self, filter_: bool, ormSession=None
    ) -> Optional[UserDetailTuple]:
        if ormSession:
            close = False

        else:
            ormSession = self._dbSessionCreator()
            close = True

        try:
            user = ormSession.query(InternalUserTuple).filter(filter_).one()
            return self._makeUserDetails(user)
        except NoResultFound:
            return None

        finally:
            if close:
                ormSession.close()

    def userByUserKeyBlocking(
        self, userKey, ormSession=None
    ) -> Optional[UserDetailTuple]:
        filter_ = InternalUserTuple.userKey == userKey
        return self._userInfoByFilter(filter_, ormSession)

    @deferToThreadWrapWithLogger(logger)
    def users(
        self,
        likeTitle: Optional[str] = None,
        groupNames: Optional[List[str]] = None,
        isFieldLogin: Optional[bool] = None,
    ):
        """
        Returns Deferred[List[UserDetailTuple]]
        """
        ormSession = self._dbSessionCreator()
        try:
            return self.usersBlocking(ormSession, likeTitle, groupNames)
        finally:
            ormSession.close()

    def usersBlocking(
        self,
        session,
        likeTitle: Optional[str] = None,
        groupNames: Optional[List[str]] = None,
        isFieldLogin: Optional[bool] = None,
    ) -> List[UserDetailTuple]:
        qry = (
            session.query(InternalUserTuple)
            .options(subqueryload(InternalUserTuple.groups))
            .order_by(InternalUserTuple.userName)
        )

        if isFieldLogin:
            qry = qry.join(
                UserLoggedIn,
                UserLoggedIn.userName == InternalUserTuple.userName,
            ).filter(UserLoggedIn.isFieldLogin)

        if groupNames:
            qry = (
                qry.join(InternalUserGroupTuple)
                .join(InternalGroupTuple)
                .filter(InternalGroupTuple.groupName.in_(groupNames))
            )

        if likeTitle:
            qry = qry.filter(
                or_(
                    InternalUserTuple.userTitle.ilike("%" + likeTitle + "%"),
                    InternalUserTuple.userName.ilike("%" + likeTitle + "%"),
                )
            )

        ormUsers = qry.all()

        return [self._makeUserDetails(u) for u in ormUsers]

    def _makeUserDetails(self, ormUser: InternalUserTuple) -> UserDetailTuple:
        userDetail = UserDetailTuple()
        for fieldName in self._userCopyFields:
            setattr(userDetail, fieldName, getattr(ormUser, fieldName))

        userDetail.groupNames = [g.groupName for g in ormUser.groups]
        return userDetail

    @deferToThreadWrapWithLogger(logger)
    def groups(self, likeTitle: Optional[str] = None):
        """
        Returns Deferred[List[UserDetailTuple]]
        """
        ormSession = self._dbSessionCreator()
        try:
            return self.groupsBlocking(ormSession, likeTitle)
        finally:
            ormSession.close()

    def groupsBlocking(
        self, session, likeTitle: Optional[str] = None
    ) -> List[GroupDetailTuple]:
        qry = session.query(InternalGroupTuple).order_by(
            InternalGroupTuple.groupName
        )

        if likeTitle:
            qry = qry.filter(
                InternalGroupTuple.userTitle.ilike("%" + likeTitle + "%")
            )

        return [self._makeGroupDetails(u) for u in qry.all()]

    def _makeGroupDetails(
        self, ormGroups: InternalGroupTuple
    ) -> GroupDetailTuple:
        groupDetail = GroupDetailTuple()
        for fieldName in self._groupCopyFields:
            setattr(groupDetail, fieldName, getattr(ormGroups, fieldName))

        return groupDetail

    @deferToThreadWrapWithLogger(logger)
    def peekDeviceTokensForUser(self, userName: str) -> List[str]:
        session = self._dbSessionCreator()
        try:
            result = (
                session.query(UserLoggedIn)
                .filter(UserLoggedIn.userName == userName)
                .all()
            )

            return [r.deviceToken for r in result]

        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def peekUserForDeviceToken(self, deviceToken) -> Optional[UserDetailTuple]:
        session = self._dbSessionCreator()
        try:
            result = (
                session.query(InternalUserTuple)
                .join(
                    UserLoggedIn,
                    UserLoggedIn.userName == InternalUserTuple.userName,
                )
                .filter(UserLoggedIn.deviceToken == deviceToken)
                .one()
            )

            return self._makeUserDetails(result)

        except NoResultFound:
            return None

        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def peekTokensWithUserDetails(
        self, isFieldDevice
    ) -> List[DeviceWithUserDetailsTuple]:
        session = self._dbSessionCreator()
        try:
            query = (
                session.query(UserLoggedIn, InternalUserTuple)
                .join(
                    InternalUserTuple,
                    UserLoggedIn.userName == InternalUserTuple.userName,
                )
                .filter(UserLoggedIn.isFieldLogin == isFieldDevice)
            )
            result = []
            for fieldLoggedInUser, internalUser in query.all():
                row = DeviceWithUserDetailsTuple(
                    deviceToken=fieldLoggedInUser.deviceToken,
                    userDetails=internalUser.toUserDetailTuple(),
                )
                result.append(row)
            return result
        except NoResultFound:
            return []
        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def peekLoggedInDeviceTokens(self, isFieldDevice) -> List[str]:
        session = self._dbSessionCreator()
        try:
            query = session.query(UserLoggedIn).filter(
                UserLoggedIn.isFieldLogin == isFieldDevice
            )
            return [row.deviceToken for row in query.all()]
        except NoResultFound:
            return []
        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def userLoggedInInfo(
        self,
        userName: Optional[str] = None,
        isFieldDevice: Optional[bool] = None,
    ) -> List[UserLoggedIn]:
        return self.userLoggedInInfoBlocking(userName, isFieldDevice)

    def userLoggedInInfoBlocking(
        self,
        userName: Optional[str] = None,
        isFieldDevice: Optional[bool] = None,
        ormSession=None,
    ) -> List[UserLoggedIn]:
        noMainThread()
        if ormSession is None:
            closeSession = True
            session = self._dbSessionCreator()
        else:
            closeSession = False
            session = ormSession

        try:
            query = session.query(UserLoggedIn)
            if userName:
                query = query.filter(UserLoggedIn.userName == userName)

            if isFieldDevice:
                query = query.filter(UserLoggedIn.isFieldLogin == isFieldDevice)

            return [row.toTuple() for row in query.all()]
        finally:
            if closeSession:
                session.close()



import logging

from sqlalchemy.orm import subqueryload

from peek_core_user._private.PluginNames import userPluginFilt
from peek_core_user._private.storage.InternalGroupTuple import (
    InternalGroupTuple,
)
from peek_core_user._private.storage.InternalUserTuple import InternalUserTuple
from peek_core_user.tuples.UserListItemTuple import UserListItemTuple
from peek_core_user.tuples.UserDetailTuple import UserDetailTuple
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler
from vortex.sqla_orm.OrmCrudHandler import (
    OrmCrudHandler,
    OrmCrudHandlerExtension,
)

from peek_core_user.tuples.constants.UserAuthTargetEnum import (
    UserAuthTargetEnum,
)

logger = logging.getLogger(__name__)

# This dict matches the definition in the Admin angular app.
filtKey = {"key": "admin.Edit.InternalUserTuple"}
filtKey.update(userPluginFilt)


# This is the CRUD hander
class __CrudHandler(OrmCrudHandler):
    def createDeclarative(self, session, payloadFilt):
        likeTitle = payloadFilt["likeTitle"]

        tuples = []

        if likeTitle and len(likeTitle) >= 3:
            tuples = (
                session.query(InternalUserTuple)
                .options(subqueryload(InternalUserTuple.groups))
                .filter(
                    InternalUserTuple.userTitle.ilike("%" + likeTitle + "%")
                )
                .all()
            )

        for tuple in tuples:
            tuple.groupIds = []
            for group in tuple.groups:
                tuple.groupIds.append(group.id)

        return tuples

    def _create(self, session, payloadFilt):
        tuples = self.createDeclarative(session, payloadFilt)
        self._ext.afterCreate(tuples, session, payloadFilt)
        return Payload(tuples=tuples).makePayloadEnvelope()


class __ExtUpdateObservable(OrmCrudHandlerExtension):
    """Update Observable ORM Crud Extension

    This extension is called after events that will alter data,
    it then notifies the observer.

    """

    def __init__(self, tupleDataObserver: TupleDataObservableHandler):
        self._tupleDataObserver = tupleDataObserver

    def _afterCommit(self, tuple_, tuples, session, payloadFilt):
        selector = {}
        # Copy any filter values into the selector
        selector["likeTitle"] = payloadFilt["likeTitle"]
        tupleSelector = TupleSelector(InternalUserTuple.tupleName(), selector)
        self._tupleDataObserver.notifyOfTupleUpdate(tupleSelector)

        self._tupleDataObserver.notifyOfTupleUpdate(
            TupleSelector(UserDetailTuple.tupleName(), {})
        )

        self._tupleDataObserver.notifyOfTupleUpdate(
            TupleSelector(UserListItemTuple.tupleName(), {})
        )

        for tuple in tuples:
            tuple.groupIds = []
            for group in tuple.groups:
                tuple.groupIds.append(group.id)

        return True

    afterUpdateCommit = _afterCommit
    afterDeleteCommit = _afterCommit

    def beforeUpdate(self, tuple_, tuples, session, payloadFilt):
        for t in tuples:
            t.userKey = t.userName.lower()
            if t.importSource is None:
                t.importSource = "PEEK_ADMIN"
                t.authenticationTarget = UserAuthTargetEnum.INTERNAL
        return True

    def middleUpdate(self, tuple_, tuples, session, payloadFilt):
        groupsById = {g.id: g for g in session.query(InternalGroupTuple)}

        for user in tuples:
            if user.groupIds:
                user.groups = [groupsById[gid] for gid in user.groupIds]

        return True


# This method creates an instance of the handler class.
def makeInternalUserTableHandler(tupleObservable, dbSessionCreator):
    handler = __CrudHandler(
        dbSessionCreator, InternalUserTuple, filtKey, retreiveAll=True
    )

    logger.debug("Started")
    handler.addExtension(
        InternalUserTuple, __ExtUpdateObservable(tupleObservable)
    )
    return handler



import logging

from vortex.Payload import Payload
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.sqla_orm.OrmCrudHandler import OrmCrudHandler

from peek_core_user._private.PluginNames import userPluginFilt
from peek_core_user._private.storage.Setting import OTP_NUMBER_OF_OPTIONS
from peek_core_user._private.storage.Setting import OTP_NUMBER_OF_WORDS
from peek_core_user._private.storage.Setting import OTP_VALID_SECONDS
from peek_core_user._private.storage.Setting import oneTimePasscodeSetting
from peek_core_user._private.tuples.OtpSettingTuple import OtpSettingTuple

logger = logging.getLogger(__name__)

# This dict matches the definition in the Admin angular app.
filtKey = {"key": "admin.Edit.OtpSetting"}
filtKey.update(userPluginFilt)


# This is the CRUD hander
class __CrudHandler(OrmCrudHandler):

    def createDeclarative(self, session, payloadFilt):
        return [self._getDeclarativeById(session, None)]

    def _getDeclarativeById(self, session, id_=None):
        settings = oneTimePasscodeSetting(session)
        return OtpSettingTuple(
            otpNumberOfWords=settings[OTP_NUMBER_OF_WORDS],
            otpNumberOfCandidates=settings[OTP_NUMBER_OF_OPTIONS],
            otpValidSeconds=settings[OTP_VALID_SECONDS],
        )

    def _update(self, session, tuples, payloadFilt) -> PayloadEnvelope:
        # Update the OTP settings in the database
        settings = oneTimePasscodeSetting(session)
        settings[OTP_NUMBER_OF_WORDS] = tuples[0].otpNumberOfWords
        settings[OTP_NUMBER_OF_OPTIONS] = tuples[0].otpNumberOfCandidates
        settings[OTP_VALID_SECONDS] = tuples[0].otpValidSeconds

        # Commit the changes
        session.commit()

        returnTuples = self._getDeclarativeById(session)
        # Return the updated settings
        return Payload(tuples=returnTuples).makePayloadEnvelope(result=True)


# This method creates an instance of the handler class.
def makeOtpSettingsHandler(tupleObservable, dbSessionCreator):
    handler = __CrudHandler(
        dbSessionCreator, OtpSettingTuple, filtKey, retreiveAll=True
    )

    return handler



"""Added vehical to LoggedIn

Peek Plugin Database Migration Script

Revision ID: 45c62f8acd87
Revises: 2c6728ebc562
Create Date: 2018-12-08 19:21:47.019204

"""

# revision identifiers, used by Alembic.
revision = "45c62f8acd87"
down_revision = "2c6728ebc562"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "UserLoggedIn",
        sa.Column("vehicle", sa.String(), nullable=True),
        schema="core_user",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("UserLoggedIn", "vehicle", schema="core_user")
    # ### end Alembic commands ###



import logging

from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from peek_core_user._private.PluginNames import userPluginTuplePrefix

logger = logging.getLogger(__name__)


@addTupleType
class OtpSettingTuple(Tuple):
    """OtpSetting

    This table stores settings to generate one-time passwords

    """

    __tupleType__ = userPluginTuplePrefix + "OtpSettingTuple"

    id: int = TupleField()
    otpNumberOfWords: int = TupleField()
    otpNumberOfCandidates: int = TupleField()
    otpValidSeconds: int = TupleField()



from abc import ABCMeta, abstractmethod

from twisted.internet.defer import Deferred


class UserAdminAuthApiABC(metaclass=ABCMeta):
    @abstractmethod
    def check(self, username: str, password: str) -> Deferred:
        """check

        :param username
        :param password

        :returns callback for success, or errback for fail
        """



from vortex.Tuple import addTupleType, TupleField, Tuple

from peek_core_user._private.PluginNames import userPluginTuplePrefix


@addTupleType
class UserLoginAuthMethodResponseTuple(Tuple):
    __tupleType__ = userPluginTuplePrefix + "UserLoginAuthMethodResponseTuple"

    AUTH_METHOD_NONE = -1
    AUTH_METHOD_OTP = 0
    AUTH_METHOD_PASSWORD = 1

    userName: str = TupleField()
    authMethod: str = TupleField(defaultValue=AUTH_METHOD_NONE)



""" 
 *  Copyright Synerty Pty Ltd 2017
 *
 *  MIT License
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""

import logging

from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import orm
from sqlalchemy.sql.schema import ForeignKey
from sqlalchemy.sql.schema import Index
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class InternalUserGroupTuple(DeclarativeBase, Tuple):
    """Internal User Group Tuple

    This table stores the relationships between the users and the groups.

    """

    __tupleType__ = userPluginTuplePrefix + "InternalUserGroupTuple"
    __tablename__ = "InternalUserGroup"

    userId = Column(
        Integer,
        ForeignKey("InternalUser.id", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )

    groupId = Column(
        Integer,
        ForeignKey("InternalGroup.id", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )

    __table_args__ = (
        Index("idx_InternalUserGroup_map", userId, groupId, unique=True),
    )

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)



from typing import Callable

from abc import ABCMeta, abstractmethod
from twisted.internet.defer import Deferred

from peek_core_user.tuples.login.UserLoginAction import UserLoginAction
from peek_core_user.tuples.login.UserLoginResponseTuple import (
    UserLoginResponseTuple,
)
from peek_core_user.tuples.login.UserLogoutAction import UserLogoutAction
from peek_core_user.tuples.login.UserLogoutResponseTuple import (
    UserLogoutResponseTuple,
)
from peek_core_user.tuples.login.UserLoginAuthMethodAction import (
    UserLoginAuthMethodAction,
)
from peek_core_user.tuples.login.UserLoginOtpAction import UserLoginOtpAction

UserPostLoginHookCallable = Callable[[UserLoginResponseTuple], Deferred]

UserPostLogoutHookCallable = Callable[[UserLogoutResponseTuple], Deferred]


class UserLoginApiABC(metaclass=ABCMeta):
    @abstractmethod
    def logout(
        self, logoutTuple: UserLogoutAction
    ) -> Deferred:  # [UserLogoutResponseTuple]:
        """Logout

        :param logoutTuple
        """

    @abstractmethod
    def login(
        self, loginTuple: UserLoginAction
    ) -> Deferred:  # [UserLoginResponseTuple]:
        """Login

        :param loginTuple
        """

    @abstractmethod
    def queryUserLoginAuthMethod(
        self, userLoginAuthMethod: UserLoginAuthMethodAction
    ) -> Deferred:
        """query for user authentication method

        :param userLoginAuthMethod
        :return: UserLoginAuthMethodResponseTuple
        """

    @abstractmethod
    def loginRequestOtp(self, userLoginOtpAction: UserLoginOtpAction):
        """Offer one-time password challenge for user authentication

        :param userLoginOtpAction
        :return: UserLoginOtpResponseTuple
        """



from txhttputil.util.ModuleUtil import filterModules


def loadPublicTuples():
    for mod in filterModules(__name__, __file__):
        __import__(mod, locals(), globals())

    from . import login
    from . import import_

    login.loadTuples()
    import_.loadTuples()



"""add lastLoginDate

Peek Plugin Database Migration Script

Revision ID: 3cbb84cc31d5
Revises: 94d5c124461d
Create Date: 2024-11-05 09:23:10.346803

"""

# revision identifiers, used by Alembic.
revision = "3cbb84cc31d5"
down_revision = "94d5c124461d"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column(
        "InternalUser",
        sa.Column("lastLoginDate", sa.DateTime(), nullable=True),
        schema="core_user",
    )
    op.add_column(
        "InternalUser",
        sa.Column("lastLoginDeviceToken", sa.String(), nullable=True),
        schema="core_user",
    )


def downgrade():
    op.drop_column("InternalUser", "lastLoginDate", schema="core_user")
    op.drop_column("InternalUser", "lastLoginDeviceToken", schema="core_user")



"""timezone aware

Peek Plugin Database Migration Script

Revision ID: 9b65be31926e
Revises: 56d805072d2a
Create Date: 2017-12-31 14:32:42.351711

"""

# revision identifiers, used by Alembic.
revision = "9b65be31926e"
down_revision = "56d805072d2a"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # core_user."UserLoggedIn"
    op.execute(
        """ALTER TABLE core_user."UserLoggedIn"
                  ALTER COLUMN "loggedInDateTime" TYPE TIMESTAMP WITH TIME ZONE
                  USING "loggedInDateTime" AT TIME ZONE 'UTC'
               """
    )


def downgrade():
    pass



"""added ldap table

Peek Plugin Database Migration Script

Revision ID: 2209cf338e6d
Revises: 5e33c6b788fd
Create Date: 2019-07-31 20:32:15.541766

"""

# revision identifiers, used by Alembic.
revision = "2209cf338e6d"
down_revision = "5e33c6b788fd"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "LdapSetting",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("ldapTitle", sa.String(), nullable=False),
        sa.Column("ldapDomain", sa.String(), nullable=False),
        sa.Column("ldapUri", sa.String(), nullable=False),
        sa.Column("ldapCNFolders", sa.String(), nullable=True),
        sa.Column("ldapOUFolders", sa.String(), nullable=True),
        sa.Column(
            "adminEnabled", sa.Boolean(), server_default="0", nullable=False
        ),
        sa.Column(
            "desktopEnabled", sa.Boolean(), server_default="0", nullable=False
        ),
        sa.Column(
            "mobileEnabled", sa.Boolean(), server_default="0", nullable=False
        ),
        sa.Column("ldapGroups", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("ldapTitle"),
        schema="core_user",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("LdapSetting", schema="core_user")
    # ### end Alembic commands ###



import logging

from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_core_user._private.PluginNames import userPluginTuplePrefix
from peek_core_user._private.storage.DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class LdapSetting(DeclarativeBase, Tuple):
    """LdapSetting

    This table stores connections and settings to LDAP servers

    """

    __tupleType__ = userPluginTuplePrefix + "LdapSettingTuple"
    __tablename__ = "LdapSetting"

    id = Column(Integer, primary_key=True, autoincrement=True)
    ldapTitle = Column(String, nullable=False, unique=True)
    ldapDomain = Column(String, nullable=False)
    ldapUri = Column(String, nullable=False)
    ldapCNFolders = Column(String, nullable=True)
    ldapOUFolders = Column(String, nullable=True)
    ldapGroups = Column(String, nullable=True)

    adminEnabled = Column(Boolean, nullable=False, server_default="0")
    desktopEnabled = Column(Boolean, nullable=False, server_default="0")
    mobileEnabled = Column(Boolean, nullable=False, server_default="0")

    agentHost = Column(String, nullable=True)

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)
